import pylab as pl
import numpy as np
import scipy as sp
import scipy.sparse.linalg.eigen.arpack as arp
from scipy import integrate
from scipy.linalg import expm
from svd_robust import svd_gesvd
from scipy.sparse.linalg import eigs
np.set_printoptions(linewidth=2000, precision=3,threshold=4000)

###############  Some useful fuctions ##########################

def svd(A):
    " SVD of A -- with the most appropriate routine"
    try:
        X, y, Z = np.linalg.svd(A,full_matrices=0)			
    except np.linalg.LinAlgError:
    	X, y, Z = svd_gesvd(A,full_matrices=0)
    return X,y,Z

def update(B,s,U,chi_max,tol):
    " Updates the B and s matrices using U and the TEBD protocol "
    L = len(B)
    d = B[0].shape[0]
    # set tol for the singular values from the outside, e.g., tol=10.**(-15)

    for k in [0,1]:
        for i_bond in range(k, L,2):
            # sweep even and odd links within the unit cell, keeping in mind periodicity
            ia = i_bond
            ib = np.mod(i_bond+1,L)
            ic = np.mod(i_bond+2,L)
            chia = B[ia].shape[1]
            chic = B[ib].shape[2]

#            print "shapes before",B[ia].shape,B[ib].shape

            # Construct theta matrix and time evolution:
            # start from B[ia] with indexes (j1,a,c) & B[ib] with indexes (j2,c,b),
            # and from U[i_bond] with indexes (i1,i2,j1,j2),
            # as well as from S[ia] with index (a):
            # at the end you should have a tensor with indexes (i2,a,i2,b),
            # thus having dimensions (d,chia,d,chic)
            theta = np.tensordot(B[ia],B[ib],axes=(2,1)) # i a j b
            theta = np.tensordot(U[i_bond],theta,axes=([2,3],[0,2])) # ip jp a b
            theta = np.tensordot(np.diag(s[ia]),theta,axes=(1,2)) # a ip jp b
            theta = np.reshape(np.transpose(theta,(1,0,2,3)),(d*chia,d*chic)) # ip a jp b

            # Schmidt deomposition, and truncation to the largest allowed bond dimension #
            X, Y, Z = svd(theta)
            chi2 = np.min([np.sum(Y>tol), chi_max])
            piv = np.zeros(len(Y), np.bool)
            piv[(np.argsort(Y)[::-1])[:chi2]] = True

            Y = Y[piv];
            X = X[:,piv] 
            Z = Z[piv,:]
        
            # Obtain the new values for B and s #
            # namely, define s[ib] as the renormalized singular values ...
            invsq = np.sqrt(sum(Y**2))
            s[ib] = Y/invsq 
        
            # and then adsorb them into X on one side, while discarding the old ones on the other,
            # i.e., multiply X from the left by s[ia]**(-1) and from the right by s[ib]
            # thus you get the new B[ia] -- B[ib] will be simply Z, modulo reshaping
            X=np.reshape(X,(d,chia,chi2))
            X = np.transpose(np.tensordot(np.diag(s[ia]**(-1)),X,axes=(1,1)),(1,0,2))
            B[ia] = np.tensordot(X, np.diag(s[ib]),axes=(2,0))
            B[ib] = np.transpose(np.reshape(Z,(chi2,d,chic)),(1,0,2))
    #        print "shapes after ",B[ia].shape,B[ib].shape

            #... and, that's it !


def entanglement_entropy(s, tol):
    " Returns the half chain entanglement entropy "
    # set tol for the Schmidt values from the outside, e.g., tol=10.**(-15)
    L = len(s)
    S=[]
    for ibond in range(L):
        x=s[ibond][s[ibond]>tol]**2
        S.append(-np.inner(np.log(x),x))
    return(S)

def entanglement_spectrum(s, tol):
    " Returns the half chain entanglement spectrum "
    # set tol for the Schmidt values from the outside, e.g., tol=10.**(-15)
    L = len(s)
    ES=[]
    for ibond in range(L):
        x=s[ibond][s[ibond]>tol]**2
        ES.append(x)
    return(ES)

def site_expectation_value(B,s,O):
    " Expectation value for a site operator "
    L=len(B)
    E=[]
    for isite in range(0,L):
        # contract the bond weights s[mod(isite-1,L)] with the tensor B[isite],
        # and with the observable O in the most appropriate way, 
        # so that in the end you are left with a 1x1x1 object
        # (squeeze is the tecnical way of converting it into a simple number) 
        sB = np.tensordot(np.diag(s[np.mod(isite-1,L)]),B[isite],axes=(1,1))
        C = np.tensordot(sB,O,axes=(1,0))
        sB=sB.conj()
        Etmp=np.tensordot(sB,C,axes=([0,1,2],[0,2,1]))
        E.append(np.squeeze(Etmp).item())
    return(E)

def bond_expectation_value(B,s,O):
    " Expectation value for a bond operator "
    L=len(B)
    E=[]
    for ibond in range(L):
        # this time contract the bond weights s[mod(ibond,L)] 
        # with the tensors B[ibond] and B[mod(ibond+1,L)]
        # and with the observable O in the most appropriate way, 
        # so that in the end you are left with a 1x1x1x1 object
        # (squeeze is the tecnical way of converting it into a simple number) 
        BB = np.tensordot(B[ibond],B[np.mod(ibond+1,L)],axes=(2,1))
        sBB = np.tensordot(np.diag(s[np.mod(ibond,L)]),BB,axes=(1,1))
        C = np.tensordot(sBB,O,axes=([1,2],[2,3]))
        sBB=np.conj(sBB)
        Etmp=np.tensordot(sBB,C,axes=([0,3,1,2],[0,1,2,3]))
        E.append(np.squeeze(Etmp).item()) 
    return E

def correlation_length(B,s):
    " Constructs the mixed transfermatrix and returns correlation length"
    L = len(B)
    chi = B[0].shape[1]

    if chi==1:
        xi=0 # special case of fully uncorrelated state
    else:
    # iteratively contract the B and conj(B) tensors:
    # the transfer matrix T should in the end be a chi**2 x chi**2 object,
    # with indexes (a,a',b,b') = (virtual left, virtual left conj, virtual right, virtual right conj)
    # glued in pairs together!
    # pay attention to the performance change when you choose different contraction orders ;-)
        T = np.tensordot(B[0],np.conj(B[0]),axes=(0,0)) #a,b,a',b'
        T = T.transpose(0,2,1,3) #a,a',b,b'
        for i in range(1,L):
            T = np.tensordot(T,B[i],axes=(2,1)) #a,a',b',i,b
            T = np.tensordot(T,np.conj(B[i]),axes=([2,3],[1,0])) #a,a',b,b'
        T = np.reshape(T,(chi**2,chi**2))

    # Obtain the ratio between the largest and 2nd largest eigenvalues
    # and the longest correlation length out of it
        eta = arp.eigs(T,k=2,which='LM',return_eigenvectors=False,ncv=20)
#        print "transf matrix eivals",eta
        xi = -L/np.log(np.min(np.abs(eta))/np.max(np.abs(eta)))
    return xi


# advanced task: write a subroutine to evaluate explicitly the correlation function
# of observables O1 and O2 at an arbitrary distance... if you feel fit with the above ;-)



def representation(G,l,R,chi):
    # Construct the mixed transfermatrix
    
    theta = np.tensordot(G[0],np.diag(l[0],0),axes=(2,0)) # i a b
    theta = np.tensordot(theta,G[1],axes=(2,1)) # i a j b
    theta = np.tensordot(theta,np.diag(l[1],0),axes=(3,0)) # i a j b

    d = R.shape[0]
    O = np.kron(R,R)
    O = np.reshape(O,[d,d,d,d])
    theta_O = np.tensordot(theta,O,axes=([0,2],[0,1])) # a b i j
    
    T = np.tensordot(theta_O,np.conj(theta),axes=([2,3],[0,2])) #a b a' b'
    T = np.transpose(T,(0,2,1,3))
    T = np.reshape(T,(chi**2,chi**2))
    
    # Obtain the dominant eigenvector
    eta,v = eigs(T,which='LM',k=1,ncv=100)
    return eta,np.reshape(v,(chi,chi))


