import itebd_functions
import itebd_pump
import pylab as pl
import numpy as np
import scipy as sp
import scipy.sparse.linalg.eigen.arpack as arp
from scipy import integrate
from scipy.linalg import expm
from svd_robust import svd_gesvd
from scipy.sparse.linalg import eigs
np.set_printoptions(linewidth=2000, precision=3,threshold=4000)

######## Define the simulation parameter ######################
d = 2  # physical dimension
L=2 # unit cell length for the simulation, 2 at least
chi_max = 50 # maximal bond dimension, typically between a few tens and some hundreds
N_init = 500 # number of cooling down steps before starting
N = 100 # number of cooling down steps to approach the new ground state
delta = 0.01 # (imaginary) time step to be used
tol= 10.**(-15) # tolerance for truncation in the bond dimension
eps=0.01 # parameter for the regularization of n_right, see lecture
Ns = 201 # number of sites for the regularization of n_right, see lecture

# intra-cell tunnelling will be fixed to 1.0 as unit of energy
t0 = 1.0 # central value of the inter-cell tunnelling in Rice-Mele model
V = 0.0 # density-density interactions on nearest neighbours

phi_list = np.arange(0,2.025,0.025)*np.pi # angle describing the pumping cycle, see lecture
# operators are defined in the initialization routines, see itebd_pump.py

n = []
du = []
dt = []
S_list = []
ES_list = []


# first initialize the MPS into a easy state, e.g., a charge-density wave
B,s = itebd_pump.init_cdw_mps(L)
n_0=itebd_pump.n_right(B,s,eps,Ns)
print "n_right (init) =",n_0

# cool down the initial state to the ground state, and measure the extra charge on the right
phi = phi_list[0]
u = [np.sin(phi),-np.sin(phi)]
t = [1.0, t0 + np.cos(phi)]
B,s = itebd_pump.tebd_rice_mele(t,u,V,N_init,B,s,chi_max,tol)
n_0=itebd_pump.n_right(B,s,eps,Ns)
print "n_right (t=0) =",n_0
#print "n_loc =",np.mean(np.abs(itebd_functions.site_expectation_value(B,s,sz)))


for phi in phi_list:
    u = [np.sin(phi),-np.sin(phi)]
    t = [1.0, t0 + np.cos(phi)]
    
    du.append(u[0]-u[1])
    dt.append(t[0]-t[1])
        
    B,s = itebd_pump.tebd_rice_mele(t,u,V,N,B,s,chi_max,tol)
# here you can play with the regularization parameters to measure n_right
    n.append(itebd_pump.n_right(B,s,eps,Ns))
#    print "n_loc =",np.mean(np.abs(itebd_functions.site_expectation_value(B,s,sz)))
    print "phi=",phi/2./np.pi,"n_right=",(n[-1]-n[0]) #%1


    S=np.mean(itebd_functions.entanglement_entropy(s,tol))
    S_list.append(S)
    ES=itebd_functions.entanglement_spectrum(s,tol)
    ES_list.append(ES)

# to do the correct plots, one should look at continuity conditions... not simply mod 1 ...
# anyway, as usual, you only care whether you are winding or not...
pl.figure(figsize=(9,3))
pl.subplot(121)
pl.plot(np.array(phi_list)/2./np.pi,(n-n[0])%1)
pl.xlabel('$\\phi/2\\pi$')
pl.ylabel('$\\Delta Q$')
pl.subplot(122)
pl.scatter(du, dt)
pl.scatter(0,0)
pl.xlabel('$\Delta u$')
pl.ylabel('$\Delta t$')
pl.show()    


ii=40 # index within g_list, i.e., which coupling are we looking at
phi=phi_list[ii]/2./np.pi
fig2 = pl.figure(figsize=[8,4])
pl.semilogy(ES_list[ii][1],'o')
pl.ylabel('$ES [\\phi= %g \\pi]$' % (phi))
pl.xlabel('$\\alpha$')
pl.show();


##### EVOLUTION IN REAL TIME -- NO ADIABATIC ASSUMPTION !!! ########

######## Define the simulation parameter ######################
chi_max = 50
N_init = 1000 # number of cooling down steps before starting
N_evol = 2500 # number of steps in the Trotter approximation
T_total = 10. # total time of the evolution, the larger the more adiabatic

delta=T_total/N_evol

V = 0.0
t0 = 1.0
    
B,s = itebd_pump.init_cdw_mps(2)
print "n_right_tmp =",itebd_pump.n_right(B,s,eps,Ns)

phi_list = np.linspace(0,2,num=N_evol+1, endpoint=True)*np.pi #np.arange(0,2.025,0.025)*np.pi

n = []
du = []
dt = []
S_list = []
ES_list = []

phi = phi_list[0]
u = [np.sin(phi),-np.sin(phi)]
t = [1.0, t0 + np.cos(phi)]
B,s = itebd_pump.tebd_rice_mele(t,u,V,N_init,B,s,chi_max,tol)
n_0=itebd_pump.n_right(B,s,eps,Ns)
print "n_right (t=0) =",n_0

for phi in phi_list:
    u = [np.sin(phi),-np.sin(phi)]
    t = [1.0, t0 + np.cos(phi)]
    #t = [t0 + np.cos(phi), 1.0]
    
    du.append(u[0]-u[1])
    dt.append(t[0]-t[1])
     
    # put a imaginary unit in front of delta in order to get evolution of the kind exp(-i delta H)
    U_bond_real,H_bond = itebd_pump.init_rice_mele_U_bond(t,u,V,1J*delta)
    itebd_functions.update(B,s,U_bond_real,chi_max,tol)
    n.append(itebd_pump.n_right(B,s,eps,Ns))

    print "phi=",phi/2./np.pi,"n_right=",n[-1]-n_0
    
    S=np.mean(itebd_functions.entanglement_entropy(s,tol))
    S_list.append(S)
    ES=itebd_functions.entanglement_spectrum(s,tol)
    ES_list.append(ES)

pl.figure(figsize=(9,4))
pl.subplot(121)
pl.plot(np.array(phi_list)/2./np.pi,n-n_0)
pl.xlabel('$\\phi/2\\pi$')
pl.ylabel('$\\Delta Q$')
pl.subplot(122)
pl.scatter(du, dt)
pl.scatter(0,0)
pl.xlabel('$\Delta u$')
pl.ylabel('$\Delta t$')
pl.show()    
